from dateroll.date import Date
from dateroll.duration import Duration
from dateroll.utils import (DATE_RANGE_HELPER_DICT, IERULE_MAPPING,
                            LASTDATEOFMONTH, datePeriodParse, fwd_or_bwd,
                            genEndtoEnd)


class Schedule:
    def __init__(
        self,
        st,
        ed,
        per="1m",
        stub="short",
        ret="l",
        monthEndRule="anniv",
        ie="[)",
        dc=NotImplementedError,
    ):
        self.stub = stub
        self.ret = ret
        self.monthEndRule = monthEndRule
        self.ie = ie
        self.dc = dc
        asof_st, dur_st = datePeriodParse(st)
        asof_ed, dur_ed = datePeriodParse(ed)
        st1 = Date(asof_st) + Duration(dur_st)
        ed1 = Date(asof_ed) + Duration(dur_ed)
        self.per_origin = per
        st1, ed1, per, bwd = fwd_or_bwd(st1, ed1, per)
        self.per = Duration(per)
        self.bwd = bwd

        if (
            st1 != ed1
        ):  # here applied ieRule; did not understand why self.st>self.ed runs a lot of code?
            ed1, st1 = (
                IERULE_MAPPING[ie](ed1, st1, Duration("1bd"))
                if "bd" in self.per.__str__().lower()
                else IERULE_MAPPING[ie](ed1, st1, Duration("1d"))
            )

        self.st = st1
        self.ed = ed1

    def __call__(self):
        if self.st == self.ed:
            st_ = self.st

            if "bd" in self.per.__str__().lower():
                st_ = st_ + Duration("0bd")
            if st_ > self.ed:
                return []
            else:
                return [st_]
        return self.genLWrapper()

    def __repr__(self):
        return f"{self.genLWrapper()}"

    def genL(self):
        """returns range of dates"""
        dts = []
        dt = self.st
        done = False
        max_iter = 365 * 100 * 4
        c = 0
        dt = dt + Duration("0bd")
        while not done:

            c += 1
            if c > max_iter:
                raise Exception("dateroll")
            dts.append(dt)

            dt += self.per
            if not (
                "d" in self.per.__str__().lower() or "w" in self.per.__str__().lower()
            ):
                dt_day = min(self.st.day, LASTDATEOFMONTH(dt).day)
                dt = dt.replace(day=dt_day)

            if not self.bwd and dt >= self.ed.datetime:
                done = True
            elif self.bwd and dt <= self.ed:
                done = True

        return dts

    def genLWrapper(self):
        dts = self.genL()
        if "bd" in str(self.per):
            dts[-1] = dts[-1] + Duration("0bd")
        lst = dts[-1] + self.per
        if lst == self.ed:

            dts.append(self.ed)
            self.stub = "full"
        else:
            if self.stub == "short":
                dts.append(self.ed)
            else:
                if self.stub != "long":
                    raise Exception(
                        "Stub either short or full, please check the stub name"
                    )

                if len(dts) % 2 == 0:
                    dts[-1] = self.ed
                else:
                    dts.append(self.ed)
        if not "d" in self.per.__str__().lower() and self.monthEndRule == "end-to-end":
            if self.st != LASTDATEOFMONTH(self.st):
                raise Exception(
                    "MonthEndRule only applies when start date is end of month"
                )
            dts = genEndtoEnd(dts)
        if "d" in self.per.__str__().lower() and "1" in self.per.__str__().lower():
            self.stub = "full"
        rs = DATE_RANGE_HELPER_DICT[self.ret.lower()](dts, self.stub)
        # handle result, if lp, it has to be period, if period<0 it has to be negative period in the result
        if not self.ret.lower() in ["ll", "df"]:
            rs = set(rs)
            print(rs)
            rs = list(sorted(rs))
        if self.ret.lower() == "df":
            f = lambda p: (
                f"-{p}" if ("-" in self.per_origin and not "-" in p) else f"{p}"
            )
            rs["dur"] = rs["dur"].apply(f)

        return rs


if __name__ == "__main__":
    ...



import datetime

import dateutil
import dateutil.relativedelta

from dateroll.calendars.calendarmath import calmath

cals = calmath.cals

period_order = (*"yhsqmwd", "cals", "roll")

PeriodLike = (dateutil.relativedelta, datetime.timedelta)

VALID_ROLL_CONVENTIONS = {
    'F',
    'P',
    'MF',
    'MP'
}


def addNones(*args, zeros=False):
    """
    helper to add None's with numbers
    None+0 = 0
    None+1 = 1
    0 + None = 0
    None+None = None (if zeros=False)
    None+None = 0 (if zeros = True)
    """
    if zeros:
        sum = 0
    else:
        sum = None
        for arg in args:
            if arg is not None:
                if not sum:
                    sum = arg
                else:
                    sum += arg
    return sum


class Duration(dateutil.relativedelta.relativedelta):
    """
    we do not inherit from relativedelta or timedelta, it is something to be considered

    Duration class represents a period of time, a duration of time, an interval of time
    i.e. some distance of time between two specific (yet unknown) dates

    there's implicitly 2 modes:

    calendar day mode - no knowledge of holidays or non-working days
        When counting days, there is no skipping over some days over others because a governing body declares that the offices are closed.

    business day mode -
        a) implicit assumption that 'WE' is always used in bd calcs
            'WE' stands for [W]eek[End] which is our internal two-letter code for the holiday of Saturdays and Sunday's during the "reasonable business period"
        b) user supplied holiday vectors can be unioned for adjusting accordingly

        bd adjustment is O(1) assuming the calenders are in memory or the unions are cached, worst
        bd adjustment is O(n) on bd's between two dates.

    A duration can have a number of:
        business days
        days (1d = 1.44bd)
        weeks  (1w=7d)
        months (1m approx 28-31d)
        years (1y=12m)

    ^ in order of seniority


    We define the "reasonable business period" (naiively) as the interval of (t-100y,t+100y)
        this could be extended: from 1582 AD to future if we have historical vectors for holiday calendar changes (i.e. juneteenth 1st occured on X, to map backwards math, and assume last forward for futuree math (e.g. last jubilee was 5cd, and happens every 50years..))
        realistically 5-7 year perfect lookup would be more than substantial for more use cases.
    """

    # on init, should we turn hemi/sem into 6mo and quarters into 3m?
    collapse_hemi_sem_quart_on_init = True

    def __init__(
        self,
        y=None,
        Y=None,
        h=None,
        H=None,
        s=None,
        S=None,
        q=None,
        Q=None,
        m=None,
        M=None,
        w=None,
        W=None,
        d=None,
        D=None,
        cals=None,
        bd=None,
        BD=None,
        roll=None,
    ):
        """
        y = year
        h = half year
        s = semester = half year
        q = quarter
        m = month
        w = week
        d = day
        bd = business days
        cals = list of 2-letter codes for calendars
        roll = roll convention (F,P,MF,MP)

        ^ non means not supplied, 0 means zero supplied

        collapse_hemi_sem_quart_on_init keeps just:
        y,m,w,d (for dateutil support)

        class instance keeps one of each, None if no value
        """
        self.y = addNones(y, Y)
        if self.collapse_hemi_sem_quart_on_init:
            _m = addNones(
                addNones(m, M),
                3 * addNones(q, Q, zeros=True),
                6 * addNones(s, S, h, H, zeros=True),
            )
            if _m != 0:
                self.m = _m
            else:
                self.m = None
        else:
            self.m = addNones(m, M)
        self.w = addNones(w, W)
        self.d = addNones(d, D)
        self.bd = addNones(bd, BD)

        if cals is not None:
            _cals = set()
            if isinstance(cals, str):
                cals = [cals]
            elif hasattr(cals, "__iter__"):
                for cal in cals:
                    if isinstance(cal, str):
                        if len(cal) == 2:
                            _cals |= {
                                cal,
                            }
                        else:
                            raise Exception(
                                f"Calendars must be 2-letter strings (not {cal})"
                            )
                    else:
                        raise Exception(
                            f"Calendars must be strings (not {type(cal).__name__})"
                        )
            self.cals = _cals
        else:
            self.cals = None

        if self.bd is None and cals and len(cals) > 0:
            self.bd = 0

        # add weekends if adjusting
        if self.bd is not None:
            if self.cals:
                self.cals |= {
                    "WE",
                }
            else:
                self.cals = {
                    "WE",
                }

        # now validate roll
        if roll is not None:
            if isinstance(roll, str):
                if roll in VALID_ROLL_CONVENTIONS:
                    self.roll = roll
                else:
                    NotImplementedError(roll)
            else:
                raise Exception("roll must be a str")
        else:
            if self.bd is not None:
                if self.bd >= 0:
                    self.roll = "F"
                elif self.bd < 0:
                    self.roll = "P"
                else:
                    raise NotImplementedError("n/a")
            else:
                self.roll = None

    @property
    def delta(self):
        """ """
        rd_args = {}
        if self.y:
            rd_args["years"] = self.y
        if self.m:
            rd_args["months"] = self.m
        if self.w:
            rd_args["weeks"] = self.w
        if self.d:
            rd_args["days"] = self.d

        rd = dateutil.relativedelta.relativedelta(**rd_args)

        return rd

    def apply_business_date_adjustment(self, from_date):
        """
        2 steps:
        1 calendar count #bd's
        2 apply roll convention
        """
        if self.bd is not None:
            adjusted = self.adjust_bds(from_date)
        else:
            adjusted = from_date
        if self.roll is not None:
            rolled_and_adjusted = self.apply_roll_convention(adjusted)
        else:
            rolled_and_adjusted = adjusted

        return rolled_and_adjusted

    def apply_roll_convention(self, from_date):
        """
        uses CalendarMath for roll
        """
        roll = self.roll
        if roll == "P":
            return calmath.prev_bd(from_date, cals=self.cals)
        elif roll == "MP":
            return calmath.prev_bd(from_date, cals=self.cals, mod=True)
        elif roll == "F":
            return calmath.next_bd(from_date, cals=self.cals)
        elif roll == "MF":
            return calmath.next_bd(from_date, cals=self.cals, mod=True)
        else:
            raise Exception("Unhandled roll: Must be /F, /P / MF/ /MP")

    def adjust_bds(self, from_date):
        """
        uses CalendarMath for bd adjustment
        """
        _d = calmath.add_bd(from_date, self.bd, cals=self.cals)
        return _d

    def simmplify(self):
        """
        excluding business days, from smallest unit (d) to largest (y)
        if units is larger than the next largest unit, and it is perfectly divisible,
        subtract equivalent units from smaller and increment the larger

        e.g.

        25mo = 2y1m (exact)
        5w = 1m1w (approx)
        28d = 1m (approx)

        make a setting to enable approx calculations

        approx:
            28-31 days is 1m
            4w is 1m
        exact:
            12m is 1y

        note - we skip 1w = exactly 7d, because no human knows what 2w3d but they know 17d is roughly 1/2 a month in their head

        note: q, s, and h are automatically converted to more senior buckets

        """
        ...

    """
    if 3m with calenda
    
    """

    @property
    def rough_days(self):
        """
        convert duration units to "days" with non-anchored period approximations
            e.g. 1y = 365.25 days
                 1bd = 365/252 days
        returns tuple (exact/approx,num days)
        that is if an approximation is used, the 1st part of the return is true
        """
        exact = True
        days = 0
        if self.y is not None:
            exact = exact and False
            days += 365.25 * self.y
        if self.m is not None:
            exact = exact and True
            days += 12 * self.m
        if self.w is not None:
            exact = exact and True
            days += 7 * self.w
        if self.d is not None:
            exact = exact and True
            days + self.d
        if self.bd is not None:
            exact = exact and False
            days += 365 / 252 * self.bd
        return exact, days

    @property
    def bd_only(self):
        if self.y is None and self.m is None and self.w is None and self.bd is not None:
            return self.bd
        else:
            return False

    def math(self, b, direction):
        """
        c = a + b
        """
        from dateroll.date.date import Date

        a = self

        # duration + duration
        if isinstance(b, Duration):
            """
            combine both
            """
            y = addNones(a.y, b.y)
            m = addNones(a.m, b.m)
            w = addNones(a.w, b.w)
            d = addNones(a.d, b.d)
            bd = addNones(a.bd, b.bd)

            # union cal sets
            # future, switch these to orNone
            if a.cals is not None:
                if b.cals is not None:
                    cals = a.cals | b.cals
                else:
                    cals = a.cals
            else:
                if b.cals is not None:
                    cals = b.cals
                else:
                    cals = None

            # roll adjustment form math (can be approx)
            # first compute diff, then roll
            """
            add error tolerlance to rough days calc, return n + epsilon
            if epsilon > tolerance, throw error11
            tol = 366/365*y+31/28*m+7/7+1/1+365/252
            see slack for c+b a/e - 8 scenarios, only 2 need approx
            
            """
            if a.roll or b.roll or a.cals or b.cals:
                abd = a.bd_only
                bbd = b.bd_only
                if abd and bbd:
                    # only bd's so EXACT diff
                    diff = abd + bbd
                else:
                    ae, adays = a.rough_days
                    be, bdays = b.rough_days
                    diff = adays + bdays
                    if not (ae and be) and diff < 0:
                        # EXACT diff failed, need to tell user approx is involved
                        print(
                            f"**Rare edge case, direction change, with bd/non-bday potential overlap. Check roll.**"
                        )

                # not if net diff is positive, rolling forwards
                if diff > 0:
                    _aroll = a.roll if a.roll is not None else ""
                    _broll = a.roll if a.roll is not None else ""
                    if "M" in _aroll or "M" in _broll:
                        roll = "MF"
                    else:
                        roll = "F"
                # if net diff is negative, rolling backwards
                else:
                    if "M" in a.roll or "M" in b.roll:
                        roll = "MP"
                    else:
                        roll = "P"
            else:
                roll = None

            c = Duration(y=y, m=m, w=w, d=d, bd=bd, roll=roll, cals=cals)
            return c

        # duration + rd() --- should not happen
        elif isinstance(b, dateutil.relativedelta.relativedelta):
            return self
            # raise NotImplementedError('need to cast rd')

        elif isinstance(b, Date):
            # adjust b for Non-bd's FIRST
            b_moved = b + self.delta * direction

            # # if you have bd's, then use bd adj and roll:
            if self.bd or self.cals or self.roll:
                shift_adj = self.apply_business_date_adjustment(b_moved)
            else:
                shift_adj = b_moved

            # convert back to dateroll.Date
            return Date.from_datetime(shift_adj)

        else:
            raise NotImplementedError

    @property
    def q(self):
        """equiv quarters"""
        raise NotImplementedError

    @property
    def s(self):
        """equiv semesters"""
        raise NotImplementedError

    ...  # q, w, h, d, bd(with cal), y {implicit act/365, can be setting,} y(with counter)
    # m, ...

    def __radd__(self, x):
        return self.math(x, 1)

    def __rsub__(self, x):
        return self.math(x, -1)

    def __add__(self, x):
        return self.math(x, 1)

    def __sub__(self, x):
        return self.math(x, -1)

    def __iadd__(self, x):
        return self.math(x, 1)

    def __isub__(self, x):
        return self.math(x, -1)

    def __repr__(self):
        """
        repr sorts units by seniority specificed in global

        future note: implicity call the simplify() method before sorting, and repr on the simpllifed version, not direct on __dict__
        consider moving ny/nm/nd/nbd to dict on class??
        """
        d = self.__dict__
        items = {k: d[k] for k in period_order if k in d}
        constructor = ""
        for k, v in self.__dict__.items():
            if v != None or (v == 0 and k == "bd"):
                if k == "roll":
                    v = f'"{v}"'
                constructor += f"{k}={v}, "
        return f'{self.__class__.__name__}({constructor.rstrip(", ")})'


PeriodLike = PeriodLike + (Duration,)


if __name__ == "__main__":
    ...



import pathlib
import pickle
import datetime

from dateutil.relativedelta import relativedelta as rd

DEFAULT_YEAR_RANGE = 200

def generate_ALL_and_WE(n=DEFAULT_YEAR_RANGE):
    """
    generate set of all days in some "big range"
    generate set of holidays where holiday is saturday or sunday in some "big range"
    return both sets
    """
    t_start = datetime.date.today() - rd(years=n)
    t_end = datetime.date.today() + rd(years=n)

    t = t_start
    ALL = set()
    WE = set()
    while t < t_end:
        ALL |= {
            t,
        }
        if t.weekday() in (5, 6):
            WE |= {
                t,
            }
        t += rd(days=1)

    return ALL, WE

def load_sample_data(cals, n=DEFAULT_YEAR_RANGE):
    ALL, WE = generate_ALL_and_WE(n)

    filename = pathlib.Path(__file__).parents[0] / 'sampledata.pkl'
    if filename.exists():
        with open(filename,'rb') as f:
            sample_data = pickle.load(f)
    else:
        raise FileNotFoundError(filename)

    sample_data.update({'ALL':ALL,'WE':WE})

    for i in ("ALL","WE","NY","LN","BR","ECB","FED"):
        if i not in cals:
            cals[i] = sample_data[i]
        else:
            print(f'[dateroll] {i} exists, skipping over.')
   
if __name__ == "__main__":
    from dateroll.calendars.calendars import Calendars

    cals = Calendars()

    load_sample_data(cals, n=200)



import datetime

from dateroll.parser.parser import parse_to_dateroll, parse_to_native

"""
need daycounters

datroll.* better than native because you can add datestrings to date objects

ddh('t') +'3m' .... it's relaly close to the base but little extra.

"""

from dateroll.calendars.calendarmath import calmath
cals = calmath.cals

class ddh:
    calmath = calmath
    cals = cals
    def __new__(self,string, convention=None):
        obj = parse_to_dateroll(string, convention=convention)
        return obj



import datetime
import pathlib
import pickle

from dateroll.calendars.calendars import Calendars

PARENT_LOCATION = pathlib.Path.home() / ".dateroll/"
PARENT_LOCATION.mkdir(exist_ok=True)
MODULE_LOCATION = PARENT_LOCATION / "calendars/"
MODULE_LOCATION.mkdir(exist_ok=True)
DATA_LOCATION_FILE = MODULE_LOCATION / "compiled_cals.db"


DEFAULT_IE = "(]"


class CalendarMath:
    """
    CalendarMath is a singleton that holds "compiled" version of your Calendars
    Calendar is a dict-on-disk object that stores vectors of all holidays.
    CalendarMath stores calendar unions (cached), bd-counter dictionaries for your calendars, and more, and has a cached-on-disk backend as well.
    If a mutation of a calendar is detected, cached is purged and CalendarMath triggers a rebuild.

    CalendarMath operationsa are used by Duration and Date:

    add_bd(d,n,cals)    - add n bd's do to d on cals
    sub_bd(d,n,cals)    - sub n bd's to d on cals
    is_bd(d,cals)       - is d a bd on cals
    diff(d1,d2,cals,ie=IE_RULE) - how many bd's between d1 and d2 on cals and using inclusion/exclusion rules (] [) () or []
    next_bd(d1,cals)    - next business day (mod=True triggers modified following)
    prev_bd(d1,cals)    - previous business day (mod=True triggered modified previous)

    calenedar unions are handled automatically and cached, if a list of calendars is provided a new set of compiled calendars is created (doesn't impact Calendar's dict-on-disk representation

    """

    def __init__(self, home=DATA_LOCATION_FILE):

        self.home = home
        self.cals = Calendars()
        self.hash = self.cals.hash
        self.ALL = self.cals["ALL"]
        self.unions = []

        self.fwd = {}
        self.bck = {}
        self.prev = {}
        self.next = {}

        self.cached_compile_all()

    def load_cache(self):
        if self.home.exists():
            file = open(self.home, "rb")
            with file:
                cached = pickle.load(file)
            return cached
        else:
            return {"hash": None}

    def save_cache(self):
        """
        cache internal state
        """
        cached = {
            "fwd": self.fwd,
            "bck": self.bck,
            "prev": self.prev,
            "next": self.next,
            "hash": self.hash,
            "unions": self.unions,
        }
        self.__dict__.update(cached)
        file = open(self.home, "wb")
        with file:
            pickle.dump(cached, file)

    def cached_compile_all(self):
        """
        cached compile/load of calenders, uses temp file in user's home for cache.
        """
        if self.home.exists():
            # check cache on disk
            if not self.has_mutated:
                # load cache
                cached = self.load_cache()
                self.__dict__.update(cached)
                return
            else:
                # cache invalidation
                print(
                    "[dateroll] calendar mutation detected, triggering re-compliation"
                )

        # compile
        self.compile_all()

    def compile_all(self):
        print("[dateroll] compiling calendars")
        cals = self.cals
        keys = cals.keys()
        for name in keys:
            dates = self.cals[name]
            self.gen_dicts(name, dates)

        self.save_cache()

    @property
    def has_mutated(self):
        cached = self.load_cache()
        return cached["hash"] != self.hash

    @property
    def recompile_if_mutated(self):
        if self.has_mutated:
            self.compile_all()

    @property
    def data_backend_present(self):
        """
        if any calendars are mutated the datemath backend is automatically removed by the Calendar class for saftey
        originally i checked "recompile_if_mutated" before calendar ops, but adds 89ms to do the unix checksome.
        This is a faster check for the existence of the data.
        """
        if not self.home.exists():
            self.recompile_if_mutated()

    def gen_dicts(self, name, dates):
        """
        generates 4 dictionaries for a given calendar
        fwd[date]->#bds from inception
        bck[#bds]->date
        prev[date]->prev bus date
        next[date]->next bus date
        """
        cal = sorted(dates)
        all = sorted(self.ALL)

        fwd = {}
        bck = {}
        prev = {}
        next = {}

        if len(cal) == 0:
            return

        last_cal = cal[0]
        last_idx = 0
        last_good = None
        for idx, dt in enumerate(all):
            if dt == last_cal:
                # is holiday
                fwd[dt] = last_idx
                bck[last_idx] = dt
                try:
                    last_cal = cal.pop(0)
                except:
                    break
            else:
                # is good bd
                last_idx += 1
                fwd[dt] = last_idx
                bck[last_idx] = dt
                prev[dt] = last_good
                last_good = dt

        self.fwd[name] = fwd
        self.bck[name] = bck
        self.prev[name] = prev
        self.next[name] = {v: k for k, v in prev.items()}

    def add_bd(self, d, n, cals, mod=False):
        """
        add's n business days on cal to d
        careful, if n==0, B should auto backtrack, need to verify
        """
        cal_name = self.union_swap(cals)

        if mod:
            raise NotImplementedError("mod")

        A = self.fwd[cal_name]
        B = self.bck[cal_name]

        from dateroll.date.date import Date

        if isinstance(d, datetime.datetime):
            d = datetime.date(d.year, d.month, d.day)
        elif isinstance(d, Date):
            d = d.date
        elif isinstance(d, datetime.date):
            pass
        else:
            raise TypeError(f"Date must be date (got {type(d).__name__})")

        bd_index = A[d]
        new_bd_index = bd_index + n
        new_dt = B[new_bd_index]

        return new_dt

    def sub_bd(self, d, n, cal, mod=False):
        """
        subtract (opposed of add)
        """
        return self.add_bd(self, d, -1 * n, cal, mod=mod)

    def is_bd(self, d, cals):
        """am i business day?"""
        self.recompile_if_mutated
        cal_name = self.union_swap(cals)
        CAL = self.fwd
        return d in CAL

    def diff(self, t1, t2, cals, ie=DEFAULT_IE):
        """
        compute business days between two dates
        """
        cal_name = self.union_swap(cals)
        print(f"[dateroll] IE rules not implemented...warning")

        fwd = self.fwd[cal_name]
        n = fwd[t2] - fwd[t1]
        return n

    def next_bd(self, d, cals, mod=False):
        """
        the next business date from d on calendars cals
        """
        if mod:
            raise NotImplementedError("mod")

        cal_name = self.union_swap(cals)
        return self.prev[cal_name][d]

    def prev_bd(self, d, cals, mod=False):
        """
        the previous business date from d on calendars cals
        """

        if mod:
            raise NotImplementedError("mod")

        cal_name = self.union_swap(cals)
        return self.next[cal_name][d]

    def union_swap(self, cals):
        """
        check calendar list or str is valid
        and if not copmiled (in self.unions) call self.union to compile the union
        """
        if isinstance(cals, str):
            return cals
        elif isinstance(cals, (tuple, list, set)):
            union_name = "u".join(cals)
            if union_name in self.fwd:
                return union_name
            else:
                self.union(cals)
                return union_name
        else:
            raise TypeError("Calendars must be a valid calendar name or list of names")

    def union(self, cals):
        """
        validate cals passed is in correct format
        union all the cals into a superset, then call gen_dicts and save_cache (no need to tell self.cals/Calendars)
        """
        if isinstance(cals, str):
            return cals
        elif isinstance(cals, (tuple, list, set)):
            union = set()
            if len(cals) == 1:
                if isinstance(cals[0], str):
                    return cals[0]
            for cal in cals:
                if not isinstance(cal, str):
                    raise TypeError(f"Calendar name must be string")
                if cal not in self.cals:
                    raise KeyError(f"There is no calendar {cal}")
                union |= set(self.cals[cal])
            union_name = "u".join(cals)
            self.unions.append(cals)
            dates = list(union)
            print(f"[dateroll] compiling new union [{union_name}]")
            self.gen_dicts(union_name, dates)
            self.save_cache()
        else:
            raise TypeError("Union must be tuple or list of cal names")

    def __repr__(self):
        '''
        Show names of cals and unions
        '''
        return f'{self.__class__.__name__}(home="{self.home}")\nCals: {self.cals.keys()}\nUnions: {self.unions}'

calmath = CalendarMath()

if __name__ == "__main__":
    calmath = CalendarMath()

    import datetime
    import time

    from dateutil.relativedelta import relativedelta as rd

    t = datetime.date(2024, 2, 25)

    a = time.time()

    print(t)
    # calmath.add_bd(t,1,cals=['WE','NY'])

    t2 = calmath.add_bd(t, 1, cals=["WE"])
    print(t2)

    t2 = calmath.add_bd(t, 100, cals=["WE"])
    print(t2)
    print(calmath.diff(t, t2, cals=["WE"]))
    t2 = t + rd(days=31)
    print(calmath.diff(t, t2, cals=["WE"]))

    b = time.time()
    print((b - a) * 1000, "ms")

    import code

    code.interact(local=locals())



import datetime
import hashlib
import os
import pathlib
import shelve

PARENT_LOCATION = pathlib.Path.home() / ".dateroll/"
PARENT_LOCATION.mkdir(exist_ok=True)
MODULE_LOCATION = PARENT_LOCATION / "calendars/"
MODULE_LOCATION.mkdir(exist_ok=True)
DATA_LOCATION_FILE = MODULE_LOCATION / "holiday_lists"

_ = shelve.open(str(DATA_LOCATION_FILE))

class Calendars:
    def __init__(self, home=str(DATA_LOCATION_FILE)):
        self.home = home

        with shelve.open(self.home)() as db:
            if "WE" not in db and "ALL" not in db:
                self.load_all_and_we()

    def keys(self):
        with shelve.open(self.home)() as db:
            return list(db.keys())

    @property
    def hash(self):
        with open(self.home.with_suffix(".db"), "rb") as f:
            return hashlib.md5(
                f.read(),
            ).hexdigest()

    def __setitem__(self, k, v):
        from dateroll.calendars.calendarmath import \
            DATA_LOCATION_FILE as calendar_math_file

        if calendar_math_file.exists():
            os.remove(calendar_math_file)

        # key must be 2-3 letter string in uppercase
        if not isinstance(k, str):
            raise Exception(f"Cal name must be string (got {type(k).__name__})")
        if len(k) < 2 or len(k) > 3:
            raise Exception(f"Cal name be 2 or 3 charts (got {len(k)})")
        if not k.isupper():
            raise Exception(f"Cal name must be all uppercase")
        # value must be a list of dates
        if not isinstance(v, (set, list, tuple)):
            raise Exception(
                f"Cal values must be a set/list/tuple (got {type(v).__name__})"
            )

        processed = []
        for i in v:
            if isinstance(i, datetime.datetime):
                dt = datetime.date(i.year, i.month, i.day)
                processed.append(dt)
            elif isinstance(i, datetime.date):
                dt = i
                processed.append(dt)
            elif hasattr(type(i),'__class__') and type(i).__class__.__name__ == 'Date':
                dt = dt.date
                processed.append(dt)
            else:
                raise Exception(
                    f"All cal dates must be of dateroll.Date or datetime.date{{time}} (got {type(i).__name__})"
                )

        with shelve.open(self.home)() as db:
            if k in db.keys():
                raise Exception(
                    f"{k} exists already, delete first.if you want to replace."
                )
            s = list(sorted(list(set(processed))))
            db[k] = s

    def __getitem__(self, k):
        return self.get(k)

    def __getattr__(self, k):
        return self.get(k)

    def __contains__(self, k):
        with shelve.open(self.home)() as db:
            return str(k) in db

    def __delitem__(self, k):
        with shelve.open(self.home)() as db:
            del db[k]

    def get(self, k):
        with shelve.open(self.home)() as db:
            return db[k]

    def clear(self):
        with shelve.open(self.home)() as db:
            db.clear()

    def __repr__(self):
        self.info
        return f'{self.__class__.__name__}(home="{self.home}.db")'

    def load_all_and_we(self):
        from dateroll.calendars.sampledata import generate_ALL_and_WE

        ALL, WE = generate_ALL_and_WE()
        self["ALL"] = ALL
        self["WE"] = WE

    def load_sample_data(self):
        from dateroll.calendars.sampledata import load_sample_data as lsd
        lsd(self)
        self.info

    @property
    def info(self):
        pattern = lambda a,b,c,d: f'{a:6}|{b:>8}|{c:12}|{d:12}'
        with shelve.open(self.home)() as db:
            stats = []
            print(pattern('name','#dates','min date','max date'))
            print(pattern('-'*6,'-'*8,'-'*12,'-'*12))
            for i in db.keys():
                l = db.get(i)
                n = len(l)
                mn = min(l)
                mx = max(l)
                print(pattern(str(i),str(n),str(mn),str(mx)))


if __name__ == "__main__":
    cals = Calendars()
    cals.load_sample_data()

    # import code;code.interact(local=locals())



import datetime

import dateutil
import dateutil.relativedelta
import dateutil.rrule

from dateroll.date.date import Date
from dateroll.duration.duration import Duration
from dateroll.schedule.schedule import Schedule
from dateroll.parser import parsers


class ParserError(Exception): ...


class Parser:
    """

    Parser input/output:

        dateroll types:

        str -> dateroll.Date
        str -> dateroll.Duration
        str -> dateutil.Schedule

        or to python native types (with flag):

        str -> datetime.date
        str -> datetime.timedelta
        str -> dateutil.relativedelta.relativedelta
        str -> dateutil.rrule.rrule


    str "string" must one of the following formats:

        1 - TodayString (today's date)
        2 - DateString (represents a date)
        3 - DurationString (represents a duration)
        4 - DateMathString (e.g. date + duration, date - date, etc...)
        5 - DateScheduleString (represents a trip of end-to-end dates)

        ^^ details on each is in the subsequent parsing sub-fucntion

    Patterns:

        TodayString

            a   TodayString                         ->   DateString

        DateString:

            b   DateString                          ->   Date | datetime.date | datetime.datetime

        DurationString:

            c   DurationString                      ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta

        DateMathString (addition subtypes:)

            d   DateString      +   Duration        ->   Date | datetime.date | datetime.datetime
            e   DurationString  +   DateString      ->   Date | datetime.date | datetime.datetime
            f   DurationString  +   DurationString  ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta
                Duration        +   DateString      ->   raise Exception
                DateString      +   DateString      ->   raise Exception

        DateMathString (subtraction subtypes:)

            g   DateString      -   DateString      ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta
            h   DateString      -   Duration        ->   Date | datetime.date | datetime.datetime
            i   DurationString  -   DurationString  ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta
                Duration        -   DateString      ->   raise Exception
                DurationString  -   DateString      ->   raise Exception

        DateScheduleString:

            j    DateString      ,   DateString      ,    Schedule | dateutil.rrule.rrule
            k    DateString      ,   DateMathString  ,    Schedule | dateutil.rrule.rrule
            l    DateMathString  ,   DateString      ,    Schedule | dateutil.rrule.rrule
                 DateMathString  ,   DateMathString  ,    Schedule | dateutil.rrule.rrule


    """

    native_date = datetime.date
    native_delta = dateutil.relativedelta.relativedelta

    def __new__(
        self, s, convention=None, use_native_types=False  # inherits from strings
    ):
        """
        Algorithm works left to right implicitly:
            1 - convert TodayStrings into DateStrings
            2 - Check for how many parts (1 or 3 allowed, "," separator)
            3 - For each part
                    3a. Match DateString's -> Date
                    3b. Match DurationString's -> Duration
                    3c. Match DateMathString -> Date or Duration
        """

        if not isinstance(s, str):
            raise ParserError("Must be string")

        if use_native_types:
            raise NotImplementedError("only dateroll types for now")

        self.convention = convention
        self.use_native_types = use_native_types

        # 1
        s1 = parsers.parseTodayString(s)

        # 2
        part = Parser.parse_maybe_many_parts(s1, convention=self.convention)
        return part

    @classmethod
    def parse_one_part(cls, untouched, convention=None):

        # 3a
        dates, nodates = parsers.parseDateString(untouched, convention=convention)
        # print('s before/after:', untouched,nodates)
        if nodates == "X":
            return dates[0]

        # 3b
        durations, nodatesordurations = parsers.parseDurationString(nodates)
        # print('s before/after:', nodates,nodatesordurations)
        if nodatesordurations == "X":
            return durations[0]
        dates_durations = dates + durations

        # 3c
        # print('before pdms',nodatesordurations)
        processed_answer = parsers.parseDateMathString(
            nodatesordurations, dates_durations
        )
        return processed_answer

    @classmethod
    def parse_maybe_many_parts(cls, s, convention=None):
        parts = s.split(",")
        if not s:
            TypeError("No empty strings")

        num_parts = len(parts)

        match num_parts:

            case 1:
                part = parts[0]
                date_or_period = cls.parse_one_part(part, convention=convention)
                return date_or_period

            case 3:
                _maybe_start, _maybe_stop, _maybe_step = parts

                maybe_start = cls.parse_one_part(_maybe_start, convention=convention)
                maybe_stop = cls.parse_one_part(_maybe_stop, convention=convention)
                maybe_step = cls.parse_one_part(_maybe_step, convention=convention)

                if isinstance(maybe_start, Date):
                    start = maybe_start
                else:
                    raise TypeError("Start of generation must be a valid Date")

                if isinstance(maybe_stop, Date):
                    stop = maybe_stop
                else:
                    raise TypeError("Stop of generation must be a valid Date")

                if isinstance(maybe_step, Duration):
                    step = maybe_step
                else:
                    raise TypeError("Step of generation must be a valid Duration")

                sch = Schedule(start=start, stop=stop, step=step)
                return sch

            case _:
                # Must
                raise Exception(
                    f"String must contain either 1 or 3 parts (not {num_parts})"
                )

        raise ParserError(f"Sorry, can't understand {s}") from None


def parse_to_native(string, convention=None):
    return Parser(string, convention=convention, use_native_types=True)


def parse_to_dateroll(string, convention=None):
    return Parser(string, convention=convention)



import datetime
import os
import pathlib
import tempfile
import unittest
import uuid
from unittest import expectedFailure

from dateroll import ddh

class TestDDH(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        ...

    @classmethod
    def tearDownClass(self):
        ...

    def testDateMathWithStr(self):
        """
        str->Date + str
        str->Date += str
        str + str->Date
        """

    def testDurationMathWithStr(self):
        """
        str->Duration + str
        str->Duration += str
        str + str->Duration

        """

    def testDateMathAdd(self):
        """
        str->Date + str->Date
        str->Date + str->Duration
        str->Duration + str->Date
        str->Duration + str->Duration
        """
    def testDateMathSub(self):
        """
        str->Date - str->Date
        str->Date - str->Duration
        str->Duration - str->Date
        str->Duration - str->Duration
        """

    def testSchedule(self):
        """
        str->Schedule (1bd,1d,1w,1m,1y)
        """


class TestsPracticalExamples(unittest.TestCase):
    def testNothing(self):
        """
        empty string
        """
        try:
            x = ddh("")
        except TypeError as e:
            assert True

    def testDate(self):
        """
        test for date strings
        """
        assert ddh("t") == datetime.date.today()
        assert ddh("5/5/5") == datetime.date(2005, 5, 5)

    def testDuration(self):
        """
        test for duration strings
        """

    def testSchedule1(self):
        """
        test for schedule strings
        """
        assert ddh("t") == datetime.date.today()


if __name__ == "__main__":
    unittest.main()

    # x = "2/24/24+3bd|WEuNY"
    # print(x, "ddh=", ddh(x))
    # # print("answer:",ddh('3/1/24'))
    # # print("answer:",ddh('t'))
    # # print("answer:",ddh('-3m|NY'))
    # # print("answer:",ddh('9m|NY+ -11m'))
    # # print("answer:",ddh('9m|NY + 11m0bd|EU'))
    # # print("answer:",ddh('-6m7y2d5q8w|NYuEUuWEuBRuMXuARuCONZ/MF'))
    # # print("answer:",ddh('9m+6m7y2d5q8w|NYuEUuWEuBRuMXuARuCONZ/MF'))
    # # print("answer:",ddh('t+1bd'))
    # # print("answer:",ddh('t-5bd'))
    # # print("answer:",ddh('t-0bd'))
    # # print("answer:",ddh('t+3m'))


import datetime
import os
import pathlib
import tempfile
import unittest
import uuid
from unittest import expectedFailure

from dateroll.calendars.calendarmath import CalendarMath


class TestStringMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename_base = (
            pathlib.Path(tempfile.gettempdir())
            / f"dateroll.testing.calmath.{uuid.uuid4()}.db"
        )
        cls.cals = CalendarMath(home=cls.filename_base)

    @classmethod
    def tearDownClass(self):
        filename = str(self.filename_base)
        os.remove(filename)
        # self.cals.cals.db.close(

    def test_addbd(self):
        """
        test add business days

        with mod:
            non-hol+0bd|WE
            hol+0bd|WE
            non-hol+0bd|WEuNY
            hol+0bd|WEuNY
            hol+1000bd
            nonhol+1000bd
        without mod:
            non-hol+0bd|WE
            hol+0bd|WE
            non-hol+0bd|WEuNY
            hol+0bd|WEuNY
            hol+1000bd
            nonhol+1000bd

        measure perf, put a threshold

        """

    def test_subbd(self):
        """
        just 1 or 2 from add, uses same mechanics

        """

    def test_isbd(self):
        """
        test is business day

        test a holiday is bd on non-union
        test a holiday in a is bd on union of a u b
        test a holiday in b is bd on union of a u b u c
        test nonholiday is not bd on non-union
        test nonholiday is not bd on union

        measure perf, put a threshold
        """

    def test_diffbd(self):
        """

        a - non-bd
        b - non-bd
        x - bd
        y - bd

        (a,b)
        (a,b]
        [a,b]
        [a,b)

        (a,y)
        (a,y]
        [a,y]
        [a,y)

        (x,b)
        (x,b]
        [x,b]
        [x,b)

        (x,y)
        (x,y]
        [x,y]
        [x,y)

        measure performance, put a threshold

        """

    def test_nextbd(self):
        """

        w/o mod

        from non-bd to bd on 1 cal
        from non-bd to bd on 2 cal union
        from bd to bd on 1 cal
        from bd to bd on 2 cal union

        w/ mod

        from non-bd to bd on 1 cal
        from non-bd to bd on 2 cal union
        from bd to bd on 1 cal
        from bd to bd on 2 cal union

        measure performance, put a threshold
        """

    def test_prevbd(self):
        """

        w/o mod

        from non-bd to bd on 1 cal
        from non-bd to bd on 2 cal union
        from bd to bd on 1 cal
        from bd to bd on 2 cal union

        w/ mod

        from non-bd to bd on 1 cal
        from non-bd to bd on 2 cal union
        from bd to bd on 1 cal
        from bd to bd on 2 cal union

        measure performance, put a threshold
        """


if __name__ == "__main__":
    unittest.main()



import re

# General patterns
OPERATOR = r" ?(\+|-)? ?"
INT_PART = r"\d+"
OPTIONAL_DECIMAL = r"(?:\.|\.\d+)?"
NUMBER = f"{INT_PART}{OPTIONAL_DECIMAL}"

# DateString Patterns (rough patterns, parser does validation)
D = M = r"\d{1,2}"
Y = r"\d{2,4}"
_ = r"(?:-|/|\A\Z)"
YMD = f"{Y}{_}{M}{_}{D}"
MDY = f"{M}{_}{D}{_}{Y}"
DMY = f"{D}{_}{M}{_}{Y}"

# DurationString atterns
PERIOD_LETTER = r"(?:bd|BD|[dDwWmMqQsShHyY])"
DATE_PERIOD = f"(?:({NUMBER})({PERIOD_LETTER}))"
REPEATING_DATE_PERIODS = f"(?:{DATE_PERIOD}{DATE_PERIOD}?{DATE_PERIOD}?{DATE_PERIOD}?{DATE_PERIOD}?)"
PIPE = r"\|"
TWOLETTERCAL = r"[A-ZA-Z][A-ZA-Z]"
REPEATING_CALUNIONS = f" ?{PIPE} ?({TWOLETTERCAL})(?:u({TWOLETTERCAL}))?(?:u({TWOLETTERCAL}))?(?:u({TWOLETTERCAL}))?(?:u({TWOLETTERCAL}))?(?:u({TWOLETTERCAL}))?(?:u({TWOLETTERCAL}))?(?:u({TWOLETTERCAL}))?"
PIPE_REPEAT_CAL_UNION = f"(?:{REPEATING_CALUNIONS})?"
ROLL = r" ?/ ?(MF|MP|F|P) ?"
PIPE_ROLL = f"(?:{ROLL})?"
RHS_PATTERN = f"{OPERATOR}{REPEATING_DATE_PERIODS} ?{PIPE_REPEAT_CAL_UNION}{PIPE_ROLL}"
MATCH_WITH_DTPERIOD = f"(.*)(?= ?){RHS_PATTERN}"
MATCH_WITHORWITHOUT_DTPERIOD = f"(?:{MATCH_WITH_DTPERIOD})|(.*)"
FULL_PATTERN = MATCH_WITHORWITHOUT_DTPERIOD
PIPE_2 = r"(\|)?"
TWOLETTERCAL_2 = r"[u]?([A-ZA-Z][A-ZA-Z])"
REPEATING_CALUNIONS_2 = f" ?(?:{PIPE_2}) ?{TWOLETTERCAL_2}{TWOLETTERCAL_2}?{TWOLETTERCAL_2}?{TWOLETTERCAL_2}?{TWOLETTERCAL_2}?{TWOLETTERCAL_2}?{TWOLETTERCAL_2}?{TWOLETTERCAL_2}?"
ROLL_2 = r" ?(/)? ?(MF|MP|F|P) ?"
PIPE_ROLL_2 = f"(?:{ROLL_2})?"
PIPE_REPEAT_CAL_UNION_2 = f"(?:{REPEATING_CALUNIONS_2})?"
COMPLETE_DURATION = f"({OPERATOR}{REPEATING_DATE_PERIODS} ?{PIPE_REPEAT_CAL_UNION_2}{PIPE_ROLL_2})"

# DateMathString Patterns
IDENT = r"^ ?(\+|-) ?X ?$"
MATH = r"^ ?((\+|-)? ?)?X(\+|-) ?X$"



from dateroll.parser import patterns
import re
import unittest

class TestPatterns(unittest.TestCase):

    def testCompileAll(self):
        for i in patterns.__dict__:
            print(i)
            pat = re.compile(i)
            assert pat is not None

if __name__ == "__main__":
    unittest.main()



from setuptools import find_packages, setup

# https://www.digitalocean.com/community/tutorials/how-to-package-and-distribute-python-applications

setup(
    name="dateroll",
    version="0.0.7",
    description="""dateroll makes working with dates less painful.""",
    author="Anthony Malizzio",
    author_email="anthony.malizzio@disent.com",
    url="https://github.com/disentcorp/dateroll",
    classifiers=(
        "Programming Language :: Python :: 3",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Utilities"
    ),
    packages=find_packages(),
    package_data={
        "dateroll": ["README.md", "LICENSE", "DOCS.md"],
        "dateroll": ["calendars/sampledata.pkl"],
    },
    include_package_data=True,
    install_requires=[
    ],
)



class Schedule: ...


# from dateroll.date import Date, DateLike
# from dateroll.period import Duration,PeriodLike
# from dateroll.utils import fwd_or_bwd
# from dateroll.utils import datePeriodParse, genEndtoEnd
# from dateroll.utils import DATE_RANGE_HELPER_DICT, IERULE_MAPPING, LASTDATEOFMONTH
# import datetime
# class Schedule:
#     def __init__(
#         self,
#         st,
#         ed,
#         per="1m",
#         stub="short",
#         ret="l",
#         monthEndRule="anniv",
#         ie="[)",
#         dc=NotImplementedError,
#     ):
#         self.stub = stub
#         self.ret = ret
#         self.monthEndRule = monthEndRule
#         self.ie = ie
#         self.dc = dc
#         asof_st, dur_st = datePeriodParse(st)
#         asof_ed, dur_ed = datePeriodParse(ed)
#         st1 = Date(asof_st) + Duration(dur_st)
#         ed1 = Date(asof_ed) + Duration(dur_ed)
#         self.per_origin = per
#         st1, ed1, per, bwd = fwd_or_bwd(st1, ed1, per)
#         self.per = Duration(per)
#         self.bwd = bwd

#         if st1 != ed1:  # here applied ieRule; did not understand why self.st>self.ed runs a lot of code?
#             ed1, st1 = IERULE_MAPPING[ie](ed1, st1, Duration("1bd")) if "bd" in self.per.__str__().lower() else IERULE_MAPPING[ie](ed1, st1, Duration("1d"))

#         self.st = st1
#         self.ed = ed1

#     def __call__(self):
#         if self.st == self.ed:
#             st_ = self.st

#             if "bd" in self.per.__str__().lower():
#                 st_ = st_ + Duration("0bd")
#             if st_ > self.ed:
#                 return []
#             else:
#                 return [st_]
#         return self.genLWrapper()

#     def __repr__(self):
#         return f"{self.genLWrapper()}"

#     def genL(self):
#         """returns range of dates"""
#         dts = []
#         dt = self.st
#         done = False
#         dt = dt + Duration("0bd")
#         while not done:
#             dt = dt.datetime if hasattr(dt,'datetime') else dt
#             dts.append(dt)

#             dt += self.per
#             if not ("d" in self.per.__str__().lower() or "w" in self.per.__str__().lower()):
#                 dt_day = min(self.st.day, LASTDATEOFMONTH(dt).day)
#                 dt = dt.replace(day=dt_day)

#             if not self.bwd and dt >= self.ed.datetime:
#                 done = True
#             elif self.bwd and dt <= self.ed:
#                 done = True

#         return dts

#     def genLWrapper(self):
#         dts = self.genL()
#         if "bd" in str(self.per):
#             dts[-1] = dts[-1] + Duration("0bd")
#         lst = dts[-1] + self.per
#         if lst == self.ed:

#             dts.append(self.ed)
#             self.stub = "full"
#         else:
#             if self.stub == "short":
#                 dts.append(self.ed)
#             else:
#                 if self.stub != "long":
#                     raise Exception("Stub either short or full, please check the stub name")

#                 if len(dts) % 2 == 0:
#                     dts[-1] = self.ed
#                 else:
#                     dts.append(self.ed)
#         if not "d" in self.per.__str__().lower() and self.monthEndRule == "end-to-end":
#             if self.st != LASTDATEOFMONTH(self.st):
#                 raise Exception("MonthEndRule only applies when start date is end of month")
#             dts = genEndtoEnd(dts)
#         if "d" in self.per.__str__().lower() and "1" in self.per.__str__().lower():
#             self.stub = "full"
#         rs = DATE_RANGE_HELPER_DICT[self.ret.lower()](dts, self.stub)
#         # handle result, if lp, it has to be period, if period<0 it has to be negative period in the result
#         if not self.ret.lower() in ["ll", "df"]:
#             rs = set(rs)
#             print(rs)
#             import code;code.interact(local=locals())
#             rs = list(sorted(rs))
#         if self.ret.lower() == "df":
#             f = lambda p: f"-{p}" if ("-" in self.per_origin and not "-" in p) else f"{p}"
#             rs["dur"] = rs["dur"].apply(f)

#         return rs


# class Schedule:
#     def __init__(self,start,stop,step,**kwargs):
#         self.days = []
#         if not isinstance(start,DateLike):
#             raise TypeError(f'Start is not date-like ({type(start).__name__})')
#         elif not isinstance(stop,DateLike):
#             raise TypeError(f'Start is not date-like ({type(start).__name__})')
#         elif not isinstance(step,Duration):
#             raise TypeError(f'Step is not period-like ({type(start).__name__})')

#         self.start = start
#         self.stop = stop
#         self.step = step

#         if start == stop:
#             self.days = [start]
#         elif start > stop:
#             temp = start
#             start = stop
#             stop = temp
#         else: # start < stop
#             d = start
#             while d < stop:
#                 self.days.append(d)
#                 d = d + self.step

#     def __len__(self):
#         return len(self.days)

#     def __repr__(self):
#         s = f"{self.__class__.__name__}({repr(self.start)},{repr(self.stop)},{repr(self.step)})"
#         return s

# if __name__ == '__main__':

#     t1 = Date('t') + Duration('1m')
#     t2 = Date('t') + Duration('3m')
#     dt = Duration('1d|WE')

#     print(t1,t2,dt)
#     s = Schedule(t1,t2,dt)
#     print(s)

#     import code;code.interact(local=locals())



import datetime
import re

from dateroll.date.date import Date
from dateroll.duration.duration import Duration
from dateroll.schedule.schedule import Schedule
from dateroll.parser import patterns

DEFAULT_CONVENTION = "american"
TODAYSTRINGVALUES = ["t", "t0", "today"]


class ParserStringsError(Exception): ...

def parseTodayString(s, convention=DEFAULT_CONVENTION):
    """
    this is [the] place where "t" is replaced
    """
    today = datetime.date.today()

    match convention:
        case "american":
            today_string = today.strftime(r"%m/%d/%Y")
        case "european":
            today_string = today.strftime(r"%d/%m/%Y")
        case "international":
            today_string = today.strftime(r"%Y/%m/%d")

    for t in TODAYSTRINGVALUES:
        if t in s:
            return s.replace(t, today_string)
    return s

def parseDateString(s, convention):
    """
    for a given convention, see if string contains 1 or 2 dates
    regex to extract string, and Date.from_string(), which calls dateutil.parser.parse

    valid DateStrings for refence:

        american: 1 or 2 digit month, 1 or 2 digit day, and 2 or 4 digit year
        european: 1 or 2 digit day, 1 or 2 digit month, and 2 or 4 digit year
        international: 2 or 4 digit year, 1 or 2 digit month, 1 or 2 digit day

    """

    if convention is None:
        convention = DEFAULT_CONVENTION

    # switch from colloquial convention to python stdlib vernacular
    match convention:
        case "american":
            pattern = patterns.MDY
            dateparser_kwargs = {}
        case "european":
            pattern = patterns.DMY
            dateparser_kwargs = {"dayfirst": True}
        case "international":
            pattern = patterns.YMD
            dateparser_kwargs = {"yearfirst ": True}
        case _:

            raise ParserStringsError("No convention provided!")

    dates = []
    matches = re.findall(pattern, s)

    if len(matches) > 2:
        raise Exception("Too many dates")

    for match in matches:
        date = Date.from_string(match, **dateparser_kwargs)
        s = s.replace(match, "X")
        dates.append(date)

    return dates, s

def process_duration_match(m: tuple):
    """
    COMPLETE_DURATION regex matches a 23-item tuple
    This function extracts the parts and calls the Duration contructor appropriately
    Starts with empty, and adds as finds from incoming tuple
    """
    # operator becomes multiplier
    duration_contructor_args = {}

    # get initial multplier (if any)
    op = m[1]
    match op:
        case "" | "+":
            mult = 1
        case "-":
            mult = -1
        case "_":
            raise Exception("Unknown operator")

    # get all the pairs
    for i in range(2, 12, 2):
        number = m[i]
        unit = m[i + 1]

        if number and unit:
            # cast number to integer
            number = int(number)
            if i < 4:
                # use multiplier on first pair
                number *= mult

            duration_contructor_args[unit] = number

    # attach calendars if any
    cals = m[13:21]
    for cal in cals:
        (
            duration_contructor_args.setdefault("cals", []).append(cal)
            if cal and cal.isupper()
            else None
        )

    # attach roll if any
    roll = m[22]
    if roll:
        duration_contructor_args["roll"] = roll
    else:
        if mult == -1:
            # if -BD is specific and no specific roll, roll should
            # should ALWAYS be P or previous business day
            duration_contructor_args["roll"] = "P"

    duration = Duration(**duration_contructor_args)
    return duration

def parseDurationString(s):
    """
    check for any DurationString:

        units:      1-9
        period's:   d,D day
                    bd,BD business day
                    w,W week
                    m,M month
                    q,Q quarter
                    y,Y year
        e.g. 1d, or 3M, or 9Y

        they can repeat: 1y3m9d = 1y + 3m + 9d

        modifiers after |
            roll convention:
                /F following
                /P previous
                /MF modified following
                /MP modified previous
            calendar as any uppercase 2 letter combo that map's to installed calendars
                WE -> by default is weekend calendar (list of all sat and sun from -100y to +100y)
                NY -> new york federal holidays
                EU -> ECB holidays
            calender unions: repreating calendars with "u" for union
                WEuNY -> all weekend holidays + NY holidays
                WUuNYuEU -> union of all 3 sets
    """
    durations = []
    matches = re.findall(patterns.COMPLETE_DURATION, s)

    if len(matches) > 2:
        raise Exception("Too many dates")

    for m in matches:
        full = m[0]
        duration = process_duration_match(m)
        # subs turn into addisions because Duration comes back with - inside
        s = s.replace(m[0], f"+X")
        durations.append(duration)

    return durations, s


def parseDateMathString(s, things):
    """
    takes only date math strings using X placeholder:
        X
        X+X
        X-X
        +X+X
        -X+X
        -X-X
    does the match..relies on items for the overload. invalid pairings will raise their own exeption.
    """

    #######
    ####### do more math, allow for repetitive patterns
    ###### use an eval context
    ##### safety is X's and valid operators
    s = s.replace(" ", "").replace("++", "+")

    ident_matches = re.match(patterns.IDENT, s)
    if ident_matches:
        if len(things) == 1:
            return things[0]
        else:
            raise NotImplementedError(f"Unhandled {s}")

    math_matches = re.match(patterns.MATH, s)
    if math_matches:
        if len(things) == 1:
            operand = things[0]

        elif len(things) == 2:
            left_hand_side = things[0]
            right_hand_side = things[1]

        if s == "X+X" or s == "+X+X":
            total = left_hand_side + right_hand_side
            return total
        elif s == "X-X" or s == "+X-X":
            total = left_hand_side - right_hand_side
            return total
    raise ParserStringsError("Cannot recognize as date math", s)


def parseScheduleString(s):
    """ """
    raise NotImplementedError



import calendar
import datetime
from datetime import timezone

import dateutil.relativedelta
import numpy
from dateutil.parser import parse

from dateroll.duration.duration import Duration

DateLike = (datetime.datetime, datetime.date)


class Date(datetime.date):
    ...

    @staticmethod
    def from_string(s, **dateparser_kwargs):
        """
        if string provided use dateutil's parser
        """
        dt = parse(s, **dateparser_kwargs)
        return Date.from_datetime(dt)

    @staticmethod
    def from_datetime(dt):
        """
        Create a Date instance from a datetime.datetime (drops time information), or datetime.date
        """
        if isinstance(dt, DateLike):
            y, m, d = dt.year, dt.month, dt.day
            return Date(y, m, d)

    def __repr__(self):
        return f'{self.__class__.__name__}("{self.strftime("%Y-%m-%d")}")'

    def isBd(self, cals=None):
        """
        am i a business day?
        """
        raise NotImplementedError

    def weekDay(self):
        """
        which day of the week am i
        """
        dt_mapping = {
            0: "Mon",
            1: "Tue",
            2: "Wed",
            3: "Thu",
            4: "Fri",
            5: "Sat",
            6: "Sun",
        }
        day_of_week = dt_mapping[self.weekday()]
        return day_of_week

    def weekMonth(self):
        """
        which week of the month, ie 1,2,3,4,5
        """
        d = numpy.array(calendar.monthcalendar(self.year, self.month))
        week_of_month = numpy.where(d == self.day)[0][0] + 1
        return week_of_month

    def weekYear(self):
        """
        week of the year, iso 8601
        """
        return self.isocalendar()[1]

    def toExcel(self):
        offset = 693594
        n = self.toordinal()
        rs = n - offset
        return rs

    def toUnix(self):
        utc_time = datetime.datetime(
            self.year, self.month, self.day, tzinfo=timezone.utc
        )
        utc_timestamp = utc_time.timestamp()
        return utc_timestamp

    @property
    def iso(self):
        return self.isoformat().split("T")[0]

    def isoStr(self):
        return self.datetime.isoformat(timespec="minutes")

    def __cmp__(self, b):
        return self.datetime == b

    @property
    def datetime(self):
        return datetime.datetime(self.year, self.month, self.day)

    @property
    def date(self):
        return datetime.date(self.year, self.month, self.day)

    @property
    def dt(self):
        return datetime.datetime(self.year, self.month, self.day)
    

    '''
    need to test add,sub,iadd,isub,radd,rsub
    '''

    # def __sub__(self, lhs):
    #     if isinstance(lhs, datetime.date) and isinstance(self, datetime.date):
    #         dt1 = lhs.toStr()
    #         dt2 = self.toStr()
    #         str_ = f"{dt2}-{dt1}"
    #         return Duration(str_)
    #     else:
    #         dt = self.strftime("%Y%m%d")
    #         dr = lhs.__str__()
    #         str_ = "".join([dt, dr])
    #         rs = Duration(dr).__sub__(self)

    #         return rs

    # def __add__(self,o):
    #     if isinstance(o,str):
    #         try:
    #             o_adj = Duration(o)
    #         except:
    #             o_adj = Date(o)
    #         res = o_adj.__add__(self)
    #         return res
    #     return super().__add__(o)

    # def __radd__(self,o):
    #     return self.__add__(o)


DateLike = DateLike + (Date,)

if __name__ == "__main__":
    ...



import datetime
import os
import pathlib
import tempfile
import unittest
import uuid
from unittest import expectedFailure

from dateroll.calendars.calendars import Calendars


class TestStringMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename_base = (
            pathlib.Path(tempfile.gettempdir())
            / f"dateroll.testing.{uuid.uuid4()}.cals"
        )
        cls.cals = Calendars(home=cls.filename_base)

    @classmethod
    def tearDownClass(self):
        filename = str(self.filename_base) + ".db"
        os.remove(filename)

    def test_keys(self):
        """
        ensure the keys method is returing a list-like object
        """
        k = self.cals.keys()
        assert isinstance(k, list)

    @expectedFailure
    def test_tests1(self):
        self.cals["aaaa"] = []

    @expectedFailure
    def test_tests2(self):
        self.cals["aa"] = []

    @expectedFailure
    def test_tests3(self):
        self.self.cals["AA"] = []

    def test_tests4(self):
        self.cals["AA"] = []

    def test_tests5(self):
        del self.cals["AA"]

    def test_tests6(self):
        self.cals["AA"] = ()

    @expectedFailure
    def test_tests7(self):
        self.cals["AA"] = (1,)

    @expectedFailure
    def test_tests8(self):
        self.cals["AA"] = (1,)

    def test_tests9(self):
        dates = [datetime.date.today()]
        self.cals["BBB"] = dates
        assert self.cals["BBB"] == dates


if __name__ == "__main__":
    unittest.main()
